=== modified file 'bzrlib/knit.py'
--- bzrlib/knit.py	2006-03-14 10:27:28 +0000
+++ bzrlib/knit.py	2006-04-07 02:01:02 +0000
@@ -66,16 +66,16 @@
 from copy import copy
 from cStringIO import StringIO
 import difflib
-import gzip
 from itertools import izip, chain
 import os
-
+import sys
 
 import bzrlib
 import bzrlib.errors as errors
 from bzrlib.errors import FileExists, NoSuchFile, KnitError, \
         InvalidRevisionId, KnitCorrupt, KnitHeaderError, \
         RevisionNotPresent, RevisionAlreadyPresent
+from bzrlib.tuned_gzip import *
 from bzrlib.trace import mutter
 from bzrlib.osutils import contains_whitespace, contains_linebreaks, \
      sha_strings
@@ -161,6 +161,10 @@
         return KnitContent(lines)
 
     def parse_line_delta_iter(self, lines):
+        for result_item in self.parse_line_delta[lines]:
+            yield result_item
+
+    def parse_line_delta(self, lines, version):
         """Convert a line based delta into internal representation.
 
         line delta is in the form of:
@@ -170,17 +174,20 @@
         internal represnetation is
         (start, end, count, [1..count tuples (revid, newline)])
         """
-        while lines:
-            header = lines.pop(0)
-            start, end, c = [int(n) for n in header.split(',')]
+        result = []
+        lines = iter(lines)
+        next = lines.next
+        # walk through the lines parsing.
+        for header in lines:
+            start, end, count = [int(n) for n in header.split(',')]
             contents = []
-            for i in range(c):
-                origin, text = lines.pop(0).split(' ', 1)
+            remaining = count
+            while remaining:
+                origin, text = next().split(' ', 1)
+                remaining -= 1
                 contents.append((origin.decode('utf-8'), text))
-            yield start, end, c, contents
-
-    def parse_line_delta(self, lines, version):
-        return list(self.parse_line_delta_iter(lines))
+            result.append((start, end, count, contents))
+        return result
 
     def lower_fulltext(self, content):
         """convert a fulltext content record into a serializable form.
@@ -192,7 +199,7 @@
     def lower_line_delta(self, delta):
         """convert a delta into a serializable form.
 
-        See parse_line_delta_iter which this inverts.
+        See parse_line_delta which this inverts.
         """
         out = []
         for start, end, c, lines in delta:
@@ -487,6 +494,12 @@
         The basis knit will be used to the largest extent possible
         since it is assumed that accesses to it is faster.
         """
+        #profile notes:
+        # 4168 calls in 14912, 2289 internal
+        # 4168 in 9711 to read_records
+        # 52554 in 1250 to get_parents
+        # 170166 in 865 to list.append
+        
         # needed_revisions holds a list of (method, version_id) of
         # versions that is needed to be fetched to construct the final
         # version of the file.
@@ -749,13 +762,20 @@
 
     def get_parents(self, version_id):
         """See VersionedFile.get_parents."""
-        self._check_versions_present([version_id])
-        return list(self._index.get_parents(version_id))
+        # perf notes:
+        # optimism counts!
+        # 52554 calls in 1264 872 internal down from 3674
+        try:
+            return self._index.get_parents(version_id)
+        except KeyError:
+            raise RevisionNotPresent(version_id, self.filename)
 
     def get_parents_with_ghosts(self, version_id):
         """See VersionedFile.get_parents."""
-        self._check_versions_present([version_id])
-        return list(self._index.get_parents_with_ghosts(version_id))
+        try:
+            return self._index.get_parents_with_ghosts(version_id)
+        except KeyError:
+            raise RevisionNotPresent(version_id, self.filename)
 
     def get_ancestry(self, versions):
         """See VersionedFile.get_ancestry."""
@@ -860,8 +880,16 @@
         # only want the _history index to reference the 1st index entry
         # for version_id
         if version_id not in self._cache:
+            index = len(self._history)
             self._history.append(version_id)
-        self._cache[version_id] = (version_id, options, pos, size, parents)
+        else:
+            index = self._cache[version_id][5]
+        self._cache[version_id] = (version_id, 
+                                   options,
+                                   pos,
+                                   size,
+                                   parents,
+                                   index)
 
     def __init__(self, transport, filename, mode, create=False):
         _KnitComponentFile.__init__(self, transport, filename, mode)
@@ -916,12 +944,16 @@
                     # index entry for version_id
                     version_id = rec[0]
                     if version_id not in self._cache:
+                        index = len(self._history)
                         self._history.append(version_id)
+                    else:
+                        index = self._cache[version_id][5]
                     self._cache[version_id] = (version_id,
                                                rec[1].split(','),
                                                int(rec[2]),
                                                int(rec[3]),
-                                               parents)
+                                               parents,
+                                               index)
                     # --- self._cache_version 
             except NoSuchFile, e:
                 if mode != 'w' or not create:
@@ -1012,13 +1044,15 @@
 
     def lookup(self, version_id):
         assert version_id in self._cache
-        return self._history.index(version_id)
+        return self._cache[version_id][5]
 
     def _version_list_to_index(self, versions):
         result_list = []
         for version in versions:
             if version in self._cache:
-                result_list.append(str(self._history.index(version)))
+                # -- inlined lookup() --
+                result_list.append(str(self._cache[version][5]))
+                # -- end lookup () --
             else:
                 result_list.append('.' + version.encode('utf-8'))
         return ' '.join(result_list)
@@ -1112,7 +1146,7 @@
                                      len(lines),
                                      digest)],
             lines,
-            ["end %s\n\n" % version_id.encode('utf-8')]))
+            ["end %s\n" % version_id.encode('utf-8')]))
         data_file.close()
         length= sio.tell()
 
@@ -1152,23 +1186,20 @@
         return df, rec
 
     def _parse_record(self, version_id, data):
+        # profiling notes:
+        # 4168 calls in 2880 217 internal
+        # 4168 calls to _parse_record_header in 2121
+        # 4168 calls to readlines in 330
         df, rec = self._parse_record_header(version_id, data)
-        lines = int(rec[2])
-        record_contents = self._read_record_contents(df, lines)
-        l = df.readline()
+        record_contents = df.readlines()
+        l = record_contents.pop()
+        assert len(record_contents) == int(rec[2])
         if l.decode('utf-8') != 'end %s\n' % version_id:
             raise KnitCorrupt(self._filename, 'unexpected version end line %r, wanted %r' 
                         % (l, version_id))
         df.close()
         return record_contents, rec[3]
 
-    def _read_record_contents(self, df, record_lines):
-        """Read and return n lines from datafile."""
-        r = []
-        for i in range(record_lines):
-            r.append(df.readline())
-        return r
-
     def read_records_iter_raw(self, records):
         """Read text records from data file and yield raw data.
 
@@ -1212,6 +1243,10 @@
         will be read in the given order.  Yields (version_id,
         contents, digest).
         """
+        # profiling notes:
+        # 60890  calls for 4168 extractions in 5045, 683 internal.
+        # 4168   calls to readv              in 1411
+        # 4168   calls to parse_record       in 2880
 
         needed_records = []
         for version_id, pos, size in records:
@@ -1229,7 +1264,7 @@
                 self._records[record_id] = (digest, content)
     
         for version_id, pos, size in records:
-            yield version_id, copy(self._records[version_id][1]), copy(self._records[version_id][0])
+            yield version_id, list(self._records[version_id][1]), self._records[version_id][0]
 
     def read_records(self, records):
         """Read records into a dictionary."""
@@ -1363,53 +1398,6 @@
 InterVersionedFile.register_optimiser(InterKnit)
 
 
-# make GzipFile faster:
-import zlib
-class GzipFile(gzip.GzipFile):
-    """Knit tuned version of GzipFile.
-
-    This is based on the following lsprof stats:
-    python 2.4 stock GzipFile write:
-    58971      0   5644.3090   2721.4730   gzip:193(write)
-    +58971     0   1159.5530   1159.5530   +<built-in method compress>
-    +176913    0    987.0320    987.0320   +<len>
-    +58971     0    423.1450    423.1450   +<zlib.crc32>
-    +58971     0    353.1060    353.1060   +<method 'write' of 'cStringIO.
-                                            StringO' objects>
-    tuned GzipFile write:
-    58971      0   4477.2590   2103.1120   bzrlib.knit:1250(write)
-    +58971     0   1297.7620   1297.7620   +<built-in method compress>
-    +58971     0    406.2160    406.2160   +<zlib.crc32>
-    +58971     0    341.9020    341.9020   +<method 'write' of 'cStringIO.
-                                            StringO' objects>
-    +58971     0    328.2670    328.2670   +<len>
-
-
-    Yes, its only 1.6 seconds, but they add up.
-    """
-
-    def write(self, data):
-        if self.mode != gzip.WRITE:
-            import errno
-            raise IOError(errno.EBADF, "write() on read-only GzipFile object")
-
-        if self.fileobj is None:
-            raise ValueError, "write() on closed GzipFile object"
-        data_len = len(data)
-        if data_len > 0:
-            self.size = self.size + data_len
-            self.crc = zlib.crc32(data, self.crc)
-            self.fileobj.write( self.compress.compress(data) )
-            self.offset += data_len
-
-    def writelines(self, lines):
-        # profiling indicated a significant overhead 
-        # calling write for each line.
-        # this batch call is a lot faster :).
-        # (4 seconds to 1 seconds for the sample upgrades I was testing).
-        self.write(''.join(lines))
-
-
 class SequenceMatcher(difflib.SequenceMatcher):
     """Knit tuned sequence matcher.
 

=== modified file 'bzrlib/repository.py'
--- bzrlib/repository.py	2006-03-20 09:58:25 +0000
+++ bzrlib/repository.py	2006-04-03 02:25:38 +0000
@@ -25,7 +25,7 @@
 import bzrlib.gpg as gpg
 from bzrlib.graph import Graph
 from bzrlib.inter import InterObject
-from bzrlib.knit import KnitVersionedFile
+from bzrlib.knit import KnitVersionedFile, KnitPlainFactory
 from bzrlib.lockable_files import LockableFiles, TransportLock
 from bzrlib.lockdir import LockDir
 from bzrlib.osutils import safe_unicode
@@ -769,7 +769,7 @@
         """
         result = Graph()
         vf = self._get_revision_vf()
-        versions = vf.versions()
+        versions = set(vf.versions())
         if not revision_ids:
             pending = set(self.all_revision_ids())
             required = set([])
@@ -784,6 +784,8 @@
                     raise errors.NoSuchRevision(self, revision_id)
                 # a ghost
                 result.add_ghost(revision_id)
+                # mark it as done so we dont try for it again.
+                done.add(revision_id)
                 continue
             parent_ids = vf.get_parents_with_ghosts(revision_id)
             for parent_id in parent_ids:
@@ -793,7 +795,7 @@
                     # no, queue it.
                     pending.add(parent_id)
             result.add_node(revision_id, parent_ids)
-            done.add(result)
+            done.add(revision_id)
         return result
 
     def _get_revision_vf(self):
@@ -1261,11 +1263,13 @@
 
     def _get_control_store(self, repo_transport, control_files):
         """Return the control store for this repository."""
-        return self._get_versioned_file_store('',
-                                              repo_transport,
-                                              control_files,
-                                              prefixed=False,
-                                              versionedfile_class=KnitVersionedFile)
+        return VersionedFileStore(
+            repo_transport,
+            prefixed=False,
+            file_mode=control_files._file_mode,
+            versionedfile_class=KnitVersionedFile,
+            versionedfile_kwargs={'factory':KnitPlainFactory()},
+            )
 
     def get_format_string(self):
         """See RepositoryFormat.get_format_string()."""
@@ -1279,7 +1283,8 @@
             file_mode = control_files._file_mode,
             prefixed=False,
             precious=True,
-            versionedfile_class=KnitVersionedFile)
+            versionedfile_class=KnitVersionedFile,
+            versionedfile_kwargs={'delta':False, 'factory':KnitPlainFactory()})
         return KnitRevisionStore(versioned_file_store)
 
     def _get_text_store(self, transport, control_files):

=== modified file 'bzrlib/store/revision/knit.py'
--- bzrlib/store/revision/knit.py	2006-03-08 11:08:06 +0000
+++ bzrlib/store/revision/knit.py	2006-03-31 08:25:41 +0000
@@ -22,7 +22,7 @@
 
 import bzrlib
 import bzrlib.errors as errors
-from bzrlib.knit import KnitVersionedFile
+from bzrlib.knit import KnitVersionedFile, KnitPlainFactory
 from bzrlib.store.revision import RevisionStore
 from bzrlib.store.versioned import VersionedFileStore
 from bzrlib.transport import get_transport
@@ -38,7 +38,8 @@
         versioned_file_store = VersionedFileStore(
             t.clone('revision-store'),
             precious=True,
-            versionedfile_class=KnitVersionedFile)
+            versionedfile_class=KnitVersionedFile,
+            versionedfile_kwargs={'delta':False, 'factory':KnitPlainFactory()})
         return KnitRevisionStore(versioned_file_store)
 
     def __str__(self):

=== modified file 'bzrlib/store/versioned/__init__.py'
--- bzrlib/store/versioned/__init__.py	2006-03-09 04:53:03 +0000
+++ bzrlib/store/versioned/__init__.py	2006-03-31 08:25:41 +0000
@@ -37,12 +37,14 @@
 
     def __init__(self, transport, prefixed=False, precious=False,
                  dir_mode=None, file_mode=None,
-                 versionedfile_class=WeaveFile):
+                 versionedfile_class=WeaveFile,
+                 versionedfile_kwargs={}):
         super(WeaveStore, self).__init__(transport,
                 dir_mode=dir_mode, file_mode=file_mode,
                 prefixed=prefixed, compressed=False)
         self._precious = precious
         self._versionedfile_class = versionedfile_class
+        self._versionedfile_kwargs = versionedfile_kwargs
 
     def _clear_cache_id(self, file_id, transaction):
         """WARNING may lead to inconsistent object references for file_id.
@@ -105,7 +107,8 @@
             #mutter("cache hit in %s for %s", self, file_id)
             return weave
         if transaction.writeable():
-            w = self._versionedfile_class(self.filename(file_id), self._transport, self._file_mode)
+            w = self._versionedfile_class(self.filename(file_id), self._transport, self._file_mode,
+                                          **self._versionedfile_kwargs)
             transaction.map.add_weave(file_id, w)
             transaction.register_dirty(w)
         else:
@@ -113,7 +116,8 @@
                                           self._transport,
                                           self._file_mode,
                                           create=False,
-                                          access_mode='r')
+                                          access_mode='r',
+                                          **self._versionedfile_kwargs)
             transaction.map.add_weave(file_id, w)
             transaction.register_clean(w, precious=self._precious)
         return w
@@ -140,14 +144,16 @@
         if self.has_id(file_id):
             self.delete(file_id, transaction)
         try:
-            weave = self._versionedfile_class(self.filename(file_id), self._transport, self._file_mode, create=True)
+            weave = self._versionedfile_class(self.filename(file_id), self._transport, self._file_mode, create=True,
+                                              **self._versionedfile_kwargs)
         except NoSuchFile:
             if not self._prefixed:
                 # unexpected error - NoSuchFile is raised on a missing dir only and that
                 # only occurs when we are prefixed.
                 raise
             self._transport.mkdir(hash_prefix(file_id), mode=self._dir_mode)
-            weave = self._versionedfile_class(self.filename(file_id), self._transport, self._file_mode, create=True)
+            weave = self._versionedfile_class(self.filename(file_id), self._transport, self._file_mode, create=True,
+                                              **self._versionedfile_kwargs)
         return weave
 
     def get_weave_or_empty(self, file_id, transaction):

=== modified file 'bzrlib/tests/test_knit.py'
--- bzrlib/tests/test_knit.py	2006-03-09 11:28:48 +0000
+++ bzrlib/tests/test_knit.py	2006-03-31 00:47:11 +0000
@@ -322,7 +322,7 @@
         instrumented_t._calls = []
         # request a last-first iteration
         results = list(k1.iter_lines_added_or_present_in_versions(['base2', 'base']))
-        self.assertEqual([('id.knit', [(0, 87), (87, 90)])], instrumented_t._calls)
+        self.assertEqual([('id.knit', [(0, 87), (87, 89)])], instrumented_t._calls)
         self.assertEqual(['text\n', 'text2\n'], results)
 
     def test_create_empty_annotated(self):

=== added file 'bzrlib/tuned_gzip.py'
--- bzrlib/tuned_gzip.py	1970-01-01 00:00:00 +0000
+++ bzrlib/tuned_gzip.py	2006-04-07 02:01:02 +0000
@@ -0,0 +1,287 @@
+# Copyright (C) 2005, 2006 by Canonical Ltd
+# Written by Robert Collins <robert.collins@canonical.com>
+#
+# This program is free software; you can redistribute it and/or modify
+# it under the terms of the GNU General Public License as published by
+# the Free Software Foundation; either version 2 of the License, or
+# (at your option) any later version.
+#
+# This program is distributed in the hope that it will be useful,
+# but WITHOUT ANY WARRANTY; without even the implied warranty of
+# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+# GNU General Public License for more details.
+#
+# You should have received a copy of the GNU General Public License
+# along with this program; if not, write to the Free Software
+# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+
+"""Bzrlib specific gzip tunings. We plan to feed these to the upstream gzip."""
+
+# make GzipFile faster:
+import gzip
+from gzip import U32, LOWU32, FEXTRA, FCOMMENT, FNAME, FHCRC
+import sys
+import struct
+import zlib
+
+__all__ = ["GzipFile"]
+
+
+class GzipFile(gzip.GzipFile):
+    """Knit tuned version of GzipFile.
+
+    This is based on the following lsprof stats:
+    python 2.4 stock GzipFile write:
+    58971      0   5644.3090   2721.4730   gzip:193(write)
+    +58971     0   1159.5530   1159.5530   +<built-in method compress>
+    +176913    0    987.0320    987.0320   +<len>
+    +58971     0    423.1450    423.1450   +<zlib.crc32>
+    +58971     0    353.1060    353.1060   +<method 'write' of 'cStringIO.
+                                            StringO' objects>
+    tuned GzipFile write:
+    58971      0   4477.2590   2103.1120   bzrlib.knit:1250(write)
+    +58971     0   1297.7620   1297.7620   +<built-in method compress>
+    +58971     0    406.2160    406.2160   +<zlib.crc32>
+    +58971     0    341.9020    341.9020   +<method 'write' of 'cStringIO.
+                                            StringO' objects>
+    +58971     0    328.2670    328.2670   +<len>
+
+
+    Yes, its only 1.6 seconds, but they add up.
+    """
+
+    def _add_read_data(self, data):
+        # 4169 calls in 183
+        # temp var for len(data) and switch to +='s.
+        # 4169 in 139
+        len_data = len(data)
+        self.crc = zlib.crc32(data, self.crc)
+        self.extrabuf += data
+        self.extrasize += len_data
+        self.size += len_data
+
+    def _read(self, size=1024):
+        # various optimisations:
+        # reduces lsprof count from 2500 to 
+        # 8337 calls in 1272, 365 internal
+        if self.fileobj is None:
+            raise EOFError, "Reached EOF"
+
+        if self._new_member:
+            # If the _new_member flag is set, we have to
+            # jump to the next member, if there is one.
+            #
+            # First, check if we're at the end of the file;
+            # if so, it's time to stop; no more members to read.
+            next_header_bytes = self.fileobj.read(10)
+            if next_header_bytes == '':
+                raise EOFError, "Reached EOF"
+
+            self._init_read()
+            self._read_gzip_header(next_header_bytes)
+            self.decompress = zlib.decompressobj(-zlib.MAX_WBITS)
+            self._new_member = False
+
+        # Read a chunk of data from the file
+        buf = self.fileobj.read(size)
+
+        # If the EOF has been reached, flush the decompression object
+        # and mark this object as finished.
+
+        if buf == "":
+            self._add_read_data(self.decompress.flush())
+            assert len(self.decompress.unused_data) >= 8, "what does flush do?"
+            self._read_eof()
+            # tell the driving read() call we have stuffed all the data
+            # in self.extrabuf
+            raise EOFError, 'Reached EOF'
+
+        self._add_read_data(self.decompress.decompress(buf))
+
+        if self.decompress.unused_data != "":
+            # Ending case: we've come to the end of a member in the file,
+            # so seek back to the start of the data for the next member which
+            # is the length of the decompress objects unused data - the first
+            # 8 bytes for the end crc and size records.
+            #
+            # so seek back to the start of the unused data, finish up
+            # this member, and read a new gzip header.
+            # (The number of bytes to seek back is the length of the unused
+            # data, minus 8 because those 8 bytes are part of this member.
+            seek_length = len (self.decompress.unused_data) - 8
+            if seek_length:
+                assert seek_length > 0
+                self.fileobj.seek(-seek_length, 1)
+
+            # Check the CRC and file size, and set the flag so we read
+            # a new member on the next call
+            self._read_eof()
+            self._new_member = True
+
+    def _read_eof(self):
+        """tuned to reduce function calls and eliminate file seeking:
+        pass 1:
+        reduces lsprof count from 800 to 288
+        4168 in 296 
+        avoid U32 call by using struct format L
+        4168 in 200
+        """
+        # We've read to the end of the file, so we should have 8 bytes of 
+        # unused data in the decompressor. If we dont, there is a corrupt file.
+        # We use these 8 bytes to calculate the CRC and the recorded file size.
+        # We then check the that the computed CRC and size of the
+        # uncompressed data matches the stored values.  Note that the size
+        # stored is the true file size mod 2**32.
+        crc32, isize = struct.unpack("<LL", self.decompress.unused_data[0:8])
+        # note that isize is unsigned - it can exceed 2GB
+        if crc32 != U32(self.crc):
+            raise IOError, "CRC check failed"
+        elif isize != LOWU32(self.size):
+            raise IOError, "Incorrect length of data produced"
+
+    def _read_gzip_header(self, bytes=None):
+        """Supply bytes if the minimum header size is already read.
+        
+        :param bytes: 10 bytes of header data.
+        """
+        """starting cost: 300 in 3998
+        15998 reads from 3998 calls
+        final cost 168
+        """
+        if bytes is None:
+            bytes = self.fileobj.read(10)
+        magic = bytes[0:2]
+        if magic != '\037\213':
+            raise IOError, 'Not a gzipped file'
+        method = ord(bytes[2:3])
+        if method != 8:
+            raise IOError, 'Unknown compression method'
+        flag = ord(bytes[3:4])
+        # modtime = self.fileobj.read(4) (bytes [4:8])
+        # extraflag = self.fileobj.read(1) (bytes[8:9])
+        # os = self.fileobj.read(1) (bytes[9:10])
+        # self.fileobj.read(6)
+
+        if flag & FEXTRA:
+            # Read & discard the extra field, if present
+            xlen = ord(self.fileobj.read(1))
+            xlen = xlen + 256*ord(self.fileobj.read(1))
+            self.fileobj.read(xlen)
+        if flag & FNAME:
+            # Read and discard a null-terminated string containing the filename
+            while True:
+                s = self.fileobj.read(1)
+                if not s or s=='\000':
+                    break
+        if flag & FCOMMENT:
+            # Read and discard a null-terminated string containing a comment
+            while True:
+                s = self.fileobj.read(1)
+                if not s or s=='\000':
+                    break
+        if flag & FHCRC:
+            self.fileobj.read(2)     # Read & discard the 16-bit header CRC
+
+    def readline(self, size=-1):
+        """Tuned to remove buffer length calls in _unread and...
+        
+        also removes multiple len(c) calls, inlines _unread,
+        total savings - lsprof 5800 to 5300
+        phase 2:
+        4168 calls in 2233
+        8176 calls to read() in 1684
+        changing the min chunk size to 200 halved all the cache misses
+        leading to a drop to:
+        4168 calls in 1977
+        4168 call to read() in 1646
+        - i.e. just reduced the function call overhead. May be worth 
+          keeping.
+        """
+        if size < 0: size = sys.maxint
+        bufs = []
+        readsize = min(200, size)    # Read from the file in small chunks
+        while True:
+            if size == 0:
+                return "".join(bufs) # Return resulting line
+
+            # c is the chunk
+            c = self.read(readsize)
+            # number of bytes read
+            len_c = len(c)
+            i = c.find('\n')
+            if size is not None:
+                # We set i=size to break out of the loop under two
+                # conditions: 1) there's no newline, and the chunk is
+                # larger than size, or 2) there is a newline, but the
+                # resulting line would be longer than 'size'.
+                if i==-1 and len_c > size: i=size-1
+                elif size <= i: i = size -1
+
+            if i >= 0 or c == '':
+                # if i>= 0 we have a newline or have triggered the above
+                # if size is not None condition.
+                # if c == '' its EOF.
+                bufs.append(c[:i+1])    # Add portion of last chunk
+                # -- inlined self._unread --
+                ## self._unread(c[i+1:], len_c - i)   # Push back rest of chunk
+                self.extrabuf = c[i+1:] + self.extrabuf
+                self.extrasize = len_c - i + self.extrasize
+                self.offset -= len_c - i
+                # -- end inlined self._unread --
+                return ''.join(bufs)    # Return resulting line
+
+            # Append chunk to list, decrease 'size',
+            bufs.append(c)
+            size = size - len_c
+            readsize = min(size, readsize * 2)
+
+    def readlines(self, sizehint=0):
+        # optimise to avoid all the buffer manipulation
+        # lsprof changed from:
+        # 4168 calls in 5472 with 32000 calls to readline()
+        # to :
+        # 4168 calls in 417.
+        # Negative numbers result in reading all the lines
+        if sizehint <= 0:
+            sizehint = -1
+        content = self.read(sizehint)
+        return content.splitlines(True)
+
+    def _unread(self, buf, len_buf=None):
+        """tuned to remove unneeded len calls.
+        
+        because this is such an inner routine in readline, and readline is
+        in many inner loops, this has been inlined into readline().
+
+        The len_buf parameter combined with the reduction in len calls dropped
+        the lsprof ms count for this routine on my test data from 800 to 200 - 
+        a 75% saving.
+        """
+        if len_buf is None:
+            len_buf = len(buf)
+        self.extrabuf = buf + self.extrabuf
+        self.extrasize = len_buf + self.extrasize
+        self.offset -= len_buf
+
+    def write(self, data):
+        if self.mode != gzip.WRITE:
+            import errno
+            raise IOError(errno.EBADF, "write() on read-only GzipFile object")
+
+        if self.fileobj is None:
+            raise ValueError, "write() on closed GzipFile object"
+        data_len = len(data)
+        if data_len > 0:
+            self.size = self.size + data_len
+            self.crc = zlib.crc32(data, self.crc)
+            self.fileobj.write( self.compress.compress(data) )
+            self.offset += data_len
+
+    def writelines(self, lines):
+        # profiling indicated a significant overhead 
+        # calling write for each line.
+        # this batch call is a lot faster :).
+        # (4 seconds to 1 seconds for the sample upgrades I was testing).
+        self.write(''.join(lines))
+
+

